#!/usr/bin/python

# The purpose of this script is to conduct the initial review of
# incoming bugs and attempt to get them into a better shape for later
# scripts to handle.
#
# We want to categorize the bugs as best we can, and request appropriate
# information so later scripts can conduct their analysis.

from arsenal_lib import *
from lpltoolkit import LaunchpadService

# Currently this script is suffering bug #336866 when trying to
# apply tags

if len(sys.argv) < 2:
    sys.stderr.write("Usage: " + sys.argv[0] + " <source-package>")
    exit(1)

opt_dryrun  = True
total_count = 0
bug_limit = 5
source_pkgs = sys.argv[1:]
lp          = LaunchpadService()
d           = lp.load_project("ubuntu")
canonical_kernel_team = lp.launchpad.people['canonical-kernel-team']
new_status  = "Incomplete"
days_old = 90

automated = """
[This is an automated message.  Apologies if it has reached you inappropriately; please just reply to this message indicating so.]
"""

# Class to abstract some of the functions of crafting a good reply
# TODO:
#   * Reduce need for indicating automated message
#   * Make the message_* stuff be registered into this class somehow
#   * Vary the message used depending on quantity of items requested
#   * Control whether to specify it's an automated message based on length of reply
#   * Divide up phrasings more cleverly, esp. responses to stuff already included
#   * Include directions for if their request is a wishlist, packaging, or workflow bug
#   * Ignore bugs starting with "Please..."
#   * Check for 'patch' in subject
#   * If a zip or gz file was uploaded in the last 3 messages, auto append request not to do that

class ArsenalReply:
    def __init__(self, bug):
        self.bug = bug
        self.message = ""

    def set_message(self, message):
        self.message = message

    def append_message(self, message):
        self.message = self.message + message

    def get_message(self):
        return self.message

    def make_reply(self):
        if self.message != "":
            greeting = "Hi " + self.bug.owner_firstname + ",\n\n"
            return greeting + self.message + automated
        else:
            return ""

# TODO:  Distinguish between rules that don't apply for certain kinds of
#        packages, like not asking for lspci for input devices, or not
#        bothering to ask for crash info for proprietary drivers, or
#        so on.
actions = [
    {
        'files':       ['backtrace', 'xorglog', 'lspci'],
        'no-files':    [],
        'symptoms':    [],
        'no-symptoms': [],
        'reply':       '',
        'status':      'Confirmed',
    },
    {
        'files':       ['xorglog', 'lspci'],
        'no-files':    ['crash'],
        'symptoms':    [],
        'no-symptoms': [],
        'reply':       '',
        'status':      'Confirmed',
    },
]

message_all = """Please be sure to confirm this issue exists with the latest development release of Ubuntu.  ISO CD images are available from http://cdimage.ubuntu.com/releases/ .  Please then run following command from a Terminal (Applications->Accessories->Terminal).  It will automatically gather and attach updated debug information to this report.

apport-collect -p %s %d

Also, if you could test the latest upstream kernel available that would be great.  It will allow additional upstream developers to examine the issue.  Refer to https://wiki.ubuntu.com/KernelMainlineBuilds .  Once you've tested the upstream kernel, please remove the 'needs-upstream-testing' tag.  This can be done by clicking on the yellow pencil icon next to the tag located at the bottom of the bug description and deleting the 'needs-upstream-testing' text.  Please let us know your results.

Thanks in advance.
"""

message_test_upstream = """If you could also please test the latest upstream kernel available that would be great.  It will allow additional upstream developers to examine the issue.  Refer to https://wiki.ubuntu.com/KernelMainlineBuilds .  Once you've tested the upstream kernel, please remove the 'needs-upstream-testing' tag.  This can be done by clicking on the yellow pencil icon next to the tag located at the bottom of the bug description and deleting the 'needs-upstream-testing' text.  Please let us know your results.

Thanks in advance.
"""

message_old_bug = """This bug was reported a while ago and there hasn't been any activity in it recently. We were wondering if this is still an issue? Can you try with the latest development release of Ubuntu?  ISO CD images are available from http://cdimage.ubuntu.com/releases/ .

If it remains an issue, could you run the following command from a Terminal (Applications->Accessories->Terminal).  It will automatically gather and attach updated debug information to this report.

apport-collect -p %s %d

Also, if you could test the latest upstream kernel available that would be great.  It will allow additional upstream developers to examine the issue.  Refer to https://wiki.ubuntu.com/KernelMainlineBuilds .  Once you've tested the upstream kernel, please remove the 'needs-upstream-testing' tag.  This can be done by clicking on the yellow pencil icon next to the tag located at the bottom of the bug description and deleting the 'needs-upstream-testing' text.  Please let us know your results.

Thanks in advance.
"""

message_sound_bug = """Please be sure to confirm this issue exists with the latest development release of Ubuntu.  ISO CD images are available from http://cdimage.ubuntu.com/releases/ . Please then run the following command from a Terminal (Applications->Accessories->Terminal).  It will automatically gather and attach updated debug information to this report.

apport-collect -p alsa-base %d

Also, if you could test the latest upstream kernel available that would be great.  It will allow additional upstream developers to examine the issue.  Refer to https://wiki.ubuntu.com/KernelMainlineBuilds .  Once you've tested the upstream kernel, please remove the 'needs-upstream-testing' tag.  This can be done by clicking on the yellow pencil icon next to the tag located at the bottom of the bug description and deleting the 'needs-upstream-testing' text.  Please let us know your results.

Thanks in advance.
"""

message_suspend_hibernate_bug ="""Please be sure to confirm this issue exists with the latest development release of Ubuntu.  ISO CD images are available from http://cdimage.ubuntu.com/releases/ .  However, note you can only test Suspend when using a LiveCD.  If you could run the following command from a Terminal (Applications->Accessories->Terminal) it will automatically gather and attach updated debug information to this report.

apport-collect -p %s %d

Also, please be sure to take a look at https://wiki.ubuntu.com/DebuggingKernelSuspendHibernateResume .  If you can provide any additional information outlined there it would be much appreciated.

Additionally, if you could try to reproduce this with the upstream mainline kernel that would be great.  It will allow additional upstream developers to examine the issue.  Refer to https://wiki.ubuntu.com/KernelMainlineBuilds .  Once you've tested the upstream kernel, please remove the 'needs-upstream-testing' tag.  This can be done by clicking on the yellow pencil icon next to the tag located at the bottom of the bug description and deleting the 'needs-upstream-testing' text.  Thanks in advance.
"""

# Table of regexes for different files
file_regexes = [
    # added via ubuntu-bug or apport-collect
    {'name':'BootDmesg',        're':'BootDmesg\.txt', },
    {'name':'CurrentDmesg',     're':'CurrentDmesg\.txt', },
    {'name':'Dependencies',     're':'Dependencies\.txt', },
    {'name':'HalComputerInfo',  're':'HalComputerInfo\.txt', },
    {'name':'LsUsb',            're':'LsUsb\.txt', },
    {'name':'Lspci',            're':'Lspci\.txt', },
    {'name':'ProcCpuInfo',      're':'ProcCpuInfo\.txt', },
    {'name':'ProcInterrupts',   're':'ProcInterrupts\.txt', },
    {'name':'ProcModules',      're':'ProcModules\.txt', },

    # added via apport-collect alsa-base
    {'name':'AlsaDevices',      're':'AlsaDevices\.txt', },
    {'name':'AplayDevices',     're':'AplayDevices\.txt', },
    {'name':'ArecordDevices',   're':'ArecordDevices\.txt', },
    {'name':'Amixer',           're':'Card0\.Amixer\.values\.txt', },
    {'name':'Codec0',          're':'Card0\.Codecs\.codec\.0\.txt', },
    {'name':'Codec1',          're':'Card0\.Codecs\.codec\.1\.txt', },
    {'name':'PciMultimedia',    're':'PciMultimedia\.txt', },

    {'name':'kernlog',    're':'kern\.log', },
    {'name':'dmesg',      're':'dmesg', },
    {'name':'lshw',       're':'lshw', },
    {'name':'lspci',      're':'lspci', },
    {'name':'lsusb',      're':'lsusb', },
    {'name':'lshal',      're':'lshal', },
    {'name':'lsmod',      're':'lsmod', },
]
for file_regex in file_regexes:
    file_regex['rc'] = re.compile(file_regex['re'], re.IGNORECASE)

# These correspond to standard ubuntu-x tags
# TODO:  Figure out how to handle these
#    * 'rotate'
#    * high memory
#    * font size/dpi
#    * hotkey issues, ' key '
#    * 'random' in title or description
#    * 'failed to install'
#    * not using .* driver
#    * 'does not boot'
symptom_regexes = [
    {'name':'error-exit',    're':'(exits with|server terminated|kills session)', },
    {'name':'error-exit',    're':'(forces a logout)', },
    {'name':'no-screens',    're':'(no screens found|screen not detected)', },
    {'name':'crash',         're':'(crash|gdm_slave_xioerror_handler)', },
    {'name':'freeze',        're':'(freeze|freezing| hangs* |locks* *up|system unresponsive)', },
    {'name':'corruption',    're':'(corrupt|rubbish|garbage|garbled|display bugs|colored noise)', },
    {'name':'tearing',       're':'(tearing)', },
    {'name':'resolution',    're':'(resolution| mode |frequency too high)', },
    {'name':'performance',   're':'(performance|sluggish|jerky|slow at redraw)', },
    {'name':'flicker',       're':'(flicker|blinking|blinks|flashes periodically)', },
    {'name':'dual-head',     're':'(dual|ext|external)[ -]*(head|monitor)', },
    {'name':'backlight',     're':'(backlight|brightness)', },
    {'name':'videoplayback', 're':'( Xv |mplayer|xine|totem)', },
    {'name':'ghost-monitor', 're':'(false|non-existant) (monitor|vga )', },
    {'name':'3d',            're':'(3D |OpenGL| GL |Blender)', },
    {'name':'compiz',        're':'( compiz|desktop effects)', },
    {'name':'resume',        're':'(suspend|resume|hibernate|wake up)', },
    {'name':'cropped',       're':'(cropped)', },
    {'name':'font-size',     're':'(text .* too large|huge font)', },
    {'name':'black-screen',  're':'(black|blank) (tv |video|screen)', },
    {'name':'black-screen',  're':'screen (goes|stays) (black|blank)', },
    {'name':'black-screen',  're':'no X.* display', },
    {'name':'vt-switch',     're':'(VT switch|ctrl-alt-f|tty switch|switch .*to console)', },
    {'name':'tv-out',        're':'(TV[ -]*Out|S[ -]*Video| HDMI )', },
    {'name':'high-cpu',      're':'(high \%*cpu|cpu \d+\%|keeps processor.*busy|high load)', },
    ]
for symptom_regex in symptom_regexes:
        symptom_regex['rc'] = re.compile(symptom_regex['re'], re.IGNORECASE)

symptoms = [
    {'name':'freeze',     'source':'xserver-xorg-video-intel', 'needs': ['xorglog', 'lspci', 'dri_debug'], },
    {'name':'crash',      'source':'any',                      'needs': ['xorglog', 'lspci', 'backtrace'], },
    {'name':'generic',    'source':'any',                      'needs': ['xorglog', 'lspci'], },
    ]

# TODO:  These should merge with the symptom/tagging stuff
regex_subj = {'workflow': re.compile('Please', re.IGNORECASE),
                'sound': re.compile('(speakers|muted|volume|sound|microphone|headphone|audio|music|alsa|pulseaudio)', re.IGNORECASE),
                'suspend': re.compile('(suspend)', re.IGNORECASE),
                'hibernate': re.compile('(hibernate)', re.IGNORECASE)}

try:
    for source_pkg in source_pkgs:
        count=0
        print "Processing ", source_pkg
        s = d.getSourcePackage(name = source_pkg)
        # Eventually update the searchTasks for tag negated search to avoid
        # checking for "kj-omit", "workflow", "review-request", and "kj-triage" below
        for bugtask in s.searchTasks(status="New"):
            if bug_limit and count >= bug_limit:
                print "Hit max number of bugs to process, exiting."
                break

            bug = ArsenalBug(bugtask.bug, lp.launchpad)
            print "https://bugs.launchpad.net/ubuntu/+bug/%d" % bug.id, "  ", bugtask.status, "  ", str(bug.title)
            reply = ArsenalReply(bug)
            logs_complete = False

            # TODO:  These bugs cause launchpad to issue a 500 error for some reason
            if bug.id == 337926 or bug.id == 377090:
                continue

            # TODO:  If a bug is nomnated, skip it
            if 0 == 1:
                continue

            # Skip bugs we don't want to handle automatically
            if (bug.has_tag("kj-omit") or
                bug.has_tag("workflow") or
                bug.has_tag("ct-rev") or
                bug.has_tag("review-request") or
                bug.has_tag("kj-triage") or
                bug.has_tag("regression-proposed") or
                bug.has_tag("regression-release") or
                bug.has_tag("regression-update") or
                bug.has_tag("regression-potential")):
                print "Skipping, tagged: ", bug.bug.tags
                continue

            # Skip bugs already assigned to canonical-kernel-team or
            # canonical-kernel-team member
            if bugtask.assignee:
                assigned = False
                if (bugtask.assignee.name == canonical_kernel_team.name):
                    print "  --> skipping, bug has an canonical-kernel-team assignee"
                    assigned = True
                else:
                    for person in canonical_kernel_team.getMembersByStatus(status="Approved"):
                        if bugtask.assignee.name == person.name:
                            print "  --> skipping, assigned to", bugtask.assignee.name
                            assigned = True
                            break
                if assigned:
                    continue

            # Skip wishlist bugs
            if bugtask.importance == "Wishlist":
                print "  --> Skipping, wishlist"
                continue

            # If bug has more than 3 bugtasks, skip it
            if len(bug.bug.bug_tasks) > 2:
                nominations = ["linux (Ubuntu Karmic)",
                "linux (Ubuntu Jaunty)",
                "linux (Ubuntu Intrepid)",
                "linux (Ubuntu Hardy)"]

                print "  --> > 2 bug tasks"
                task_cnt = 1
                for task in bug.bug.bug_tasks:
                    if (("(Ubuntu" not in task.bug_target_name) or
                        (task.status in ["Invalid", "Won't Fix", "Fix Released"])):
                        print "  Skipping,", task.bug_target_name, task.status
                        continue
                    elif task.bug_target_name in nominations:
                        print "  --> nominated: ", task.bug_target_name
                        print "  --> tag review-request"
                        if not opt_dryrun:
                            bug.append_tag("review-request")
                        continue
                    elif task.bug_target_name != source_pkg + " (Ubuntu)":
                        task_cnt += 1
                        print "task_cnt: ", task_cnt, task.bug_target_name

                print "task_cnt:", task_cnt
                if task_cnt > 2:
                    print "  --> tag review-request"
                    if not opt_dryrun:
                        bug.append_tag("review-request")
                    count += 1
                    total_count += 1
                    continue

            if opt_dryrun:
                print "== Dry Run =="

            if bug.has_tag("apport-package") or bug.has_tag("apport-kerneloops"):
                logs_complete = True

            # TODO:  Move message construction logic into ArsenalReply
            for key in regex_subj.keys():
                if regex_subj[key].search(bug.title):
                    if opt_dryrun:
                        print "  --> tag", key
                    else:
                        bug.append_tag("kernel-%s" % (key))
                    continue

            # TODO:  Copy tags from https://wiki.ubuntu.com/X/Tagging
            has_symptom = {
                'crash':     0,
            }

            # Analyze title and description for common issues, then set tags accordingly
            # TODO:  If no symptoms found in title, perhaps try description??
            for symptom in symptom_regexes:
                # TODO:  Test this with dryrun set first
                continue
                if symptom['name'] in bug.bug.tags:
                    has_symptom[symptom['name']] = 1
                    print "  --> Already tagged ", symptom['name']

                elif symptom['rc'].search(title):
                    if not opt_dryrun:
                        bug.append_tag(tag['name'])
                    print "  --> Tagged symptom ", symptom['name']
                    tagged += 1

            # TODO:  Generate this automatically (or is it even necessary to pre-seed it?)
            has_file = {
                # added via ubuntu-bug or apport-collect
                'BootDmesg':    0,
                'CurrentDmesg':    0,
                'Dependencies':    0,
                'HalComputerInfo':    0,
                'LsUsb':    0,
                'Lspci':    0,
                'ProcCpuInfo':    0,
                'ProcInterrupts':    0,
                'ProcModules':    0,

                # added via apport-collect alsa-base
                'AlsaDevices':    0,
                'AplayDevices':    0,
                'ArecordDevices':    0,
                'Amixer':    0,
                'Codec0':    0,
                'Codec1':    0,
                'PciMultimedia':    0,

                'kernlog':    0,
                'dmesg':    0,
                'lshw':    0,
                'lspci':    0,
                'lsusb':    0,
                'lshal':    0,
                'lsmod':    0,
                'unknown':   0,
            }

            if bug.age_last_message() > days_old and not bug.has_tag("apport-package"):
                print "  --> Old bug over ", bug.age_last_message(), " days"
                # After Karmic, s/linux-image-`uname -r`/source_pkg/
                reply.set_message(message_old_bug % ("linux-image-`uname -r`", bug.id))
                if opt_dryrun:
                    print "  --> tag needs-kernel-logs"
                    print "  --> tag needs-upstream-testing"
                else:
                    bug.append_tag("needs-kernel-logs")
                    bug.append_tag("needs-upstream-testing")

            elif bug.has_tag("apport-bug"):
                logs_complete = True
                reply.set_message(message_test_upstream)
                if opt_dryrun:
                    print "  --> tag needs-upstream-testing"
                else:
                    bug.append_tag("needs-upstream-testing")

            elif len(bug.attachments) == 0:
                print "  --> [No attachments]"
                # After Karmic, s/linux-image-`uname -r`/source_pkg/
                if regex_subj['sound'].search(bug.title):
                    reply.set_message(message_sound_bug % (bug.id))
                elif regex_subj['suspend'].search(bug.title) or regex_subj['hibernate'].search(bug.title):
                    reply.set_message(message_suspend_hibernate_bug % ("linux-image-`uname -r`", bug.id))
                else:
                    reply.set_message(message_all % ("linux-image-`uname -r`", bug.id))

                if opt_dryrun:
                    print "  --> tag needs-kernel-logs"
                    print "  --> tag needs-upstream-testing"
                else:
                    bug.append_tag("needs-kernel-logs")
                    bug.append_tag("needs-upstream-testing")

            elif not logs_complete:
                # Analyze what sorts of files are attached
                for a in bug.attachments:
                    hosted_file = a.data
                    # The attachment content is downloaded in full during the call to
                    # open(), so this can be slow.
                    hosted_file_buffer = hosted_file.open()
                    t = hosted_file_buffer.filename
                    #content = hosted_file_buffer.read()

                    # TODO: Is a patch attached?  If so, set to Confirmed directly; maybe assign me?

                    is_unknown = True
                    if a.message.owner == bug.owner:
                        for file_regex in file_regexes:
                            if not file_regex['rc'].search(t):
                                continue

                        # Flag the hit
                            print "  --> %s [%s]" % (t, file_regex['name'])
                            has_file[file_regex['name']] = 1
                            is_unknown = False
                            break

                        if is_unknown:
                            print "  --> Unknown file"
                            has_file['unknown'] = 1

                    # TODO:  Search bug comments and Xorg.0.log's for lspci info, and if not, ask for it

                if not regex_subj['sound'].search(bug.title):
                    if ((has_file['dmesg'] or has_file['BootDmesg']or has_file['kernlog']) and (has_file['lspci'] or has_file['Lspci'])):
                        logs_complete = True
                    else:
                        if regex_subj['suspend'].search(bug.title) or regex_subj['hibernate'].search(bug.title):
                            reply.set_message(message_suspend_hibernate_bug % ("linux-image-`uname -r`", bug.id))
                        else:
                            reply.set_message(message_all % ("linux-image-`uname -r`", bug.id))
                else:
                    if (has_file['Amixer'] and has_file['AplayDevice'] and has_file['Codec0']) or has_file['unknown']:
                        logs_complete = True
                        if not bug.has_tag("kernel-sound"):
                            if opt_dryrun:
                                print "  --> tag kernel-sound"
                            else:
                                bug.append_tag("kernel-sound")
                    else:
                        reply.set_message(message_sound_bug % (bug.id))

                if not logs_complete:
                    if opt_dryrun:
                        print "  --> tag needs-kernel-logs"
                        print "  --> tag needs-upstream-testing"
                    else:
                        bug.append_tag("needs-kernel-logs")
                        bug.append_tag("needs-upstream-testing")


# TODO:  This inadvertantly results in lspci being asked twice
#            if not has_file['lspci']:
#                reply.append_message(message_also_lspci_vvnn)
#                bug.append_tag("needs-lspci-vvnn")

# TODO:  Differentiate between crashes with and without full backtraces
#            if has_symptom['crash']:
#                reply.append_message(message_backtrace)

            # Try to identify packaging bugs
            # Subject =~ "depend on"|"wishlist"|"^Please"|"needs"

            if opt_dryrun:
                print "  --> tag kj-triage"
            else:
                bug.append_tag("kj-triage")

            if reply.get_message():
                text = ""
                try:
                    text = reply.make_reply()
                    if opt_dryrun:
                        print "  --> Message was: ", text
                    else:
                        bug.append_comment(text)
                except:
                    print " ERROR:  Failed to update bug ", bug.id
                    print "Message was:  ", text
                    raise

            if logs_complete and bug.age_last_message() <= days_old :
                if opt_dryrun:
                    print "  --> Confirmed"
                else:
                    try:
                        bugtask.transitionToStatus(status = "Confirmed")
                        print "  --> Confirmed"
                    except:
                        print " ERROR:  Failed to update Status for bug ", bug.id
                        print "New status was:  ", new_status
                        raise
            else:
                if opt_dryrun:
                    print "  --> ", new_status
                else:
                    bugtask.transitionToStatus(status = new_status)
            count += 1
            total_count += 1

            # TODO:  Need to test this in a dryrun first
            if 1 == 2:
                # TODO:  Turn this into a subroutine
                for action in actions:
                    do_action = True
                    for f in actions['files']:
                        if not has_file[f]:
                            do_action = False
                            break
                    for nf in actions['no-files']:
                        if has_file[nf]:
                            do_action = False
                            break
                    for s in actions['symptoms']:
                        if not has_symptom[s]:
                            do_action = False
                            break
                    for ns in actions['non-symptoms']:
                        if has_symptom[ns]:
                            do_action = False
                            break

                    if do_action:
                        if action['status']:
                            print " --> %s (has %s)" % (action['status'], 'needed files') # TODO: list files
                        if action['reply']:
                            print "Would give reply here"  # TODO

        print count, " bugs updated for ", source_pkg

except:
    print total_count, " bugs updated"
    raise

print total_count, " bugs updated"
