#!/usr/bin/python

# The purpose of this script is to conduct the initial review of
# incoming bugs and attempt to get them into a better shape for later
# scripts to handle.
#
# We want to categorize the bugs as best we can, and request appropriate
# information so later scripts can conduct their analysis.  We also want
# to kick out bug reports that aren't valid for handling by the X team.

from arsenal_lib import *
from lpltoolkit import *

# Currently this script is suffering bug #336866 when trying to
# apply tags

if len(sys.argv) < 2:
    sys.stderr.write("Usage: " + sys.argv[0] + " <source-package>")
    exit(1)

opt_dryrun  = 0
total_count = 0
source_pkgs = sys.argv[1:]
lp          = LaunchpadService()
d           = lp.load_project("ubuntu")
new_status  = "Incomplete"

#arsenal.Debug(1)

automated = """
[This is an automated message.  Apologies if it has reached you inappropriately; please just reply to this message indicating so.]
"""

# Class to abstract some of the functions of crafting a good reply
# TODO:
#   * Reduce need for indicating automated message
#   * Make the message_* stuff be registered into this class somehow
#   * Vary the message used depending on quantity of items requested
#   * Control whether to specify it's an automated message based on length of reply
#   * Divide up phrasings more cleverly, esp. responses to stuff already included
#   * Include directions for if their request is a wishlist, packaging, or workflow bug
#   * Ignore bugs starting with "Please..."
#   * Check for 'patch' in subject
#   * If a zip or gz file was uploaded in the last 3 messages, auto append request not to do that, except for failsafe X tarballs
#   * If resolution mentioned, request xrandr output
class ArsenalReply:
    def __init__(self, bug):
        self.bug = bug
        self.message = ""

    def set_message(self, message):
        self.message = message

    def append_message(self, message):
        self.message = self.message + message

    def get_message(self):
        return self.message

    def make_reply(self):
        if self.message != "":
            greeting = "Hi " + self.bug.owner_firstname + ",\n\n"
            return greeting + self.message + automated
        else:
            return ""

# TODO:  Distinguish between rules that don't apply for certain kinds of
#        packages, like not asking for lspci for input devices, or not
#        bothering to ask for crash info for proprietary drivers, or
#        so on.
actions = [
    {
        'files':       ['backtrace', 'xorglog', 'lspci'],
        'no-files':    [],
        'symptoms':    [],
        'no-symptoms': [],
        'reply':       '',
        'status':      'Confirmed',
    },
    {
        'files':       ['xorglog', 'lspci'],
        'no-files':    ['crash'],
        'symptoms':    [],
        'no-symptoms': [],
        'reply':       '',
        'status':      'Confirmed',
    },
]

message_all = """
Please attach the output of `lspci -vvnn` and `dmesg`, and attach your /var/log/Xorg.0.log (and maybe Xorg.0.log.old) file from after reproducing this issue.  If you're using a custom /etc/X11/xorg.conf please attach that as well.
"""
message_img_only = """
Thanks for including an image to demonstrate the issue.  Could you also please attach the output of `lspci -vvnn` and `dmesg`, and attach your /var/log/Xorg.0.log or Xorg.0.log.old file from after reproducing this issue.  If you've using a custom /etc/X11/xorg.conf please attach that as well.
"""
message_some = """
Thanks for including the attached files.  Could you also include your /var/log/Xorg.0.log (or Xorg.0.log.old) from after reproducing the issue?
"""
message_backtrace = """
Could you please collect a full backtrace from when this crash occurs?  Directions on collecting full backtraces are at https://wiki.ubuntu.com/X/Backtracing
"""
message_also_xorglog = """
Also, could you please attach your /var/log/Xorg.0.log from after reproducing the issue?
"""
message_also_lspci_vvnn = """
Please attach the output of `lspci -vvnn` too.
"""
message_xkeyboard_config = """
Please attach the output of 'setxkbmap -print' and 'xkbcomp :0 -'.
"""
message_old_bug = """
Thank you for taking the time to report this bug and helping to make Ubuntu better. You reported this bug a while ago and there hasn't been any activity in it recently. We were wondering is this still an issue for you? Can you try with the latest development release of Ubuntu?  (ISO CD images are available from http://cdimage.ubuntu.com/releases/)

If it remains an issue, could you also attach a new /var/log/Xorg.0.log?
Thanks in advance.
"""

# Table of regexes for different files
file_regexes = [
    {'name':'system_env', 're':'system', },

    {'name':'image',      're':'(png|jpg)', },
    {'name':'xorglog',    're':'xorg.*log', },
    {'name':'gdmlog',     're':':.*log', },
    {'name':'xorgconf',   're':'xorg.*conf', },
    {'name':'xrandr',     're':'xrandr', },
    {'name':'glxinfo',    're':'glxinfo', },
    {'name':'setxkbmap',  're':'setxkbmap', },
    {'name':'monitors',   're':'monitors\.xml', },

    {'name':'strace',     're':'strace', },
    {'name':'backtrace',  're':'ktrace', },
    {'name':'dri_debug',  're':'dri_debug.*tgz', },
    {'name':'vmstat',     're':'vmstat', },
    {'name':'coredump',   're':'coredump', },

    {'name':'deps',       're':'Dependencies\.txt', },
    {'name':'ProcMaps',   're':'ProcMaps\.txt', },
    {'name':'ProcStatus', 're':'ProcStatus\.txt', },
    {'name':'disassembly','re':'Disassembly\.txt', },
    {'name':'registers',  're':'registers', },
    {'name':'dpkg',       're':'dpkg\.log', },
    {'name':'hal',        're':'HalComputerInfo\.txt', },
    {'name':'cpuinfo',    're':'ProcCpuinfo\.txt', },
    {'name':'interrupts', 're':'ProcInterrupts\.txt', },
    {'name':'modules',    're':'ProcModules\.txt', },
    {'name':'dmesg',      're':'dmesg\.txt', },

    {'name':'lshw',       're':'lshw', },
    {'name':'lspci',      're':'lspci', },
    {'name':'lsusb',      're':'lsusb', },
    {'name':'lshal',      're':'lshal', },
    {'name':'lsmod',      're':'lsmod', },
]
for file_regex in file_regexes:
    file_regex['rc'] = re.compile(file_regex['re'], re.IGNORECASE)

# These correspond to standard ubuntu-x tags
# TODO:  Figure out how to handle these:
#    * 'rotate'
#    * high memory
#    * font size/dpi
#    * hotkey issues, ' key '
#    * 'random' in title or description
#    * 'failed to install'
#    * not using .* driver
#    * 'does not boot'
#    * 'random'
#    * kde/kubuntu/konqueror/kwin
symptom_regexes = [
    {'name':'error-exit',    're':'(exits with|server terminated|kills session)', },
    {'name':'error-exit',    're':'(forces a logout)', },
    {'name':'no-screens',    're':'(no screens found|screen not detected)', },
    {'name':'crash',         're':'(crash|gdm_slave_xioerror_handler)', },
    {'name':'freeze',        're':'(freeze|freezing| hangs* |locks* *up|system unresponsive)', },
    {'name':'corruption',    're':'(corrupt|rubbish|garbage|garbled|display bugs|colored noise)', },
    {'name':'tearing',       're':'(tearing)', },
    {'name':'resolution',    're':'(resolution| mode |frequency too high)', },
    {'name':'performance',   're':'(performance|sluggish|jerky|slow at redraw)', },
    {'name':'flicker',       're':'(flicker|blinking|blinks|flashes periodically)', },
    {'name':'dual-head',     're':'(dual|ext|external)[ -]*(head|monitor)', },
    {'name':'backlight',     're':'(backlight|brightness)', },
    {'name':'videoplayback', 're':'( Xv |mplayer|xine|totem)', },
    {'name':'ghost-monitor', 're':'(false|non-existant) (monitor|vga )', },
    {'name':'3d',            're':'(3D |OpenGL| GL |Blender)', },
    {'name':'compiz',        're':'( compiz|desktop effects)', },
    {'name':'resume',        're':'(suspend|resume|hibernate|wake up)', },
    {'name':'cropped',       're':'(cropped)', },
    {'name':'font-size',     're':'(text .* too large|huge font)', },
    {'name':'black-screen',  're':'(black|blank) (tv |video|screen)', },
    {'name':'black-screen',  're':'screen (goes|stays) (black|blank)', },
    {'name':'black-screen',  're':'no X.* display', },
    {'name':'vt-switch',     're':'(VT switch|ctrl-alt-f|tty switch|switch .*to console)', },
    {'name':'tv-out',        're':'(TV[ -]*Out|S[ -]*Video| HDMI )', },
    {'name':'high-cpu',      're':'(high \%*cpu|cpu \d+\%|keeps processor.*busy|high load)', },
    ]
for symptom_regex in symptom_regexes:
        symptom_regex['rc'] = re.compile(symptom_regex['re'], re.IGNORECASE)

symptoms = [
    {'name':'freeze',     'source':'xserver-xorg-video-intel', 'needs': ['xorglog', 'lspci', 'dri_debug'], },
    {'name':'crash',      'source':'any',                      'needs': ['xorglog', 'lspci', 'backtrace'], },
    {'name':'generic',    'source':'any',                      'needs': ['xorglog', 'lspci'], },
    ]


# TODO:  These should merge with the symptom/tagging stuff
regex_subj_app_crash   = re.compile('(\w+) crashed with SIGSEGV')
regex_subj_crash       = re.compile('(crash|SIGSEGV)',     re.IGNORECASE)
regex_subj_workflow    = re.compile('(Please sync|Please merge|debian/control)', re.IGNORECASE)

regex_desc_crash    = re.compile('ProblemType: Crash')
regex_desc_xorglog  = re.compile('X.Org X Server')

try:
    for source_pkg in source_pkgs:
        count=0
        if opt_dryrun:
            print "\nProcessing ", source_pkg
        s = d.getSourcePackage(name = source_pkg)
        for bugtask in s.searchTasks(status="New"):
            bug = ArsenalBug(bugtask.bug, lp.launchpad)
            bug_title = bug.title
            bug_description = bug.description
            reply = ArsenalReply(bug)

            # Skip bugs we don't want to handle automatically
            if bug.has_tag("omit") or bug.has_tag("kubuntu") or bug.has_tag("ppc") or bug.has_tag("checkbox-bug"):
                continue

            # Skip wishlist bugs
            if bugtask.importance == "Wishlist":
                continue

            # Don't try to handle bugs that have been automatically processed before
            if bug.has_text_match("This is an automated message"):
                continue

            # TODO: Skip bugs that look like sync requests
            # Subject =~ /Debian/ or has ubuntu-main-sponsors subscribed

            # TODO: Identify if bug is filed against jaunty,karmic,etc. and tag accordingly

            # If bug has more than 3 open ubuntu bugtasks, skip it
            num_tasks = 0
            for task in bug.bug.bug_tasks:
                if not task.is_complete:
                    num_tasks += 1
            if num_tasks > 3:
                #print "  --> Too many open bug tasks.  Needs manual review"
                continue

            # TODO:  Copy tags from https://wiki.ubuntu.com/X/Tagging
            has_symptom = {
                'crash':     0,
            }

            # TODO:  Move message construction logic into ArsenalReply
            if regex_subj_workflow.search(bug.title):
                if not opt_dryrun:
                    bug.append_tag("workflow")
                print bug.id, "  " + bugtask.status + "  " + bug.title + "  [Workflow]"
                continue

            # Analyze title and description for common issues, then set tags accordingly
            # TODO:  If no symptoms found in title, perhaps try description??
            if source_pkg == 'xserver-xorg-video-intel' or source_pkg == 'xserver-xorg-video-ati':
                for symptom in symptom_regexes:
                    # TODO:  Test this with dryrun set first
                    continue

                    if symptom['name'] in bug.bug.tags:
                        has_symptom[symptom['name']] = 1
                        print "  --> Already tagged ", symptom['name']

                    elif symptom['rc'].search(title):
                        if not dryrun:
                            bug.append_tag(tag['name'])
                        print "  --> Tagged symptom ", symptom['name']
                        tagged += 1

            # TODO:  Also check if a backtrace is shown -> 'crash'
            # TODO:  This prints out before the bug.id stuff
            if regex_subj_crash.search(bug.title) or \
                    regex_desc_crash.search(bug.description):
                has_symptom['crash'] = 1
                bug.append_tag("crash")

            # Check if bug's description is too short
            

            # TODO:  Generate this automatically (or is it even necessary to pre-seed it?)
            has_file = {
                'image':     0,
                'xorglog':   0,
                'gdmlog':    0,
                'conf':      0,
                'strace':    0,
                'backtrace': 0,
                'lspci':     0,
                'deps':      0,
                'unknown':   0,
                'dmesg':     0,
            }

            if regex_desc_xorglog.search(bug.description):
                has_file['xorglog'] = 1

            if len(bug.attachments) == 0:
                print bug.id, "  ", bugtask.status, "  ", bug.title, "  [No attachments]"
                if not has_file['xorglog']:
                    reply.set_message(message_all)
                    bug.append_tag("needs-xorglog")
                    bug.append_tag("needs-lspci-vvnn")
            else:
                # Analyze what sorts of files are attached
                for a in bug.attachments:
                    try:
                        hosted_file = a.data
                        # The attachment content is downloaded in full during the call to
                        # open(), so this can be slow.
                        hosted_file_buffer = hosted_file.open()
                        t = hosted_file_buffer.filename
                        #content = hosted_file_buffer.read()

                        # TODO: Is a patch attached?  If so, set to Confirmed directly; maybe assign me?

                        is_unknown = True
                        for file_regex in file_regexes:
                            if not file_regex['rc'].search(t):
                                continue
                        
                            # Flag the hit
                            dbg("  %s [%s]" % (t, file_regex['name']))
                            has_file[file_regex['name']] = 1
                            is_unknown = False
                            break

                        if is_unknown:
                            has_file['unknown'] = 1
                        # TODO:  Search bug comments and Xorg.0.log's for lspci info, and if not, ask for it
                        # TODO:  For intel, I need dmidecode

                    except:
                        print "*** Exception *** encountered while checking attachment contents"

                if not has_file['xorglog'] and not has_file['unknown']:
                    bug.append_tag("needs-xorglog")
                    if has_file['image']:
                        reply.set_message(message_img_only)
                    elif not has_file['lspci']:
                        reply.set_message(message_some + message_also_lspci_vvnn)
                    else:
                        reply.set_message(message_some)


            if bug.age_last_message() > 120:
                print bug.id, "  ", bugtask.status, "  ", bug.title
                print " + Old bug over ", bug.age_last_message(), " days"
                reply.set_message(message_old_bug)
                bug.append_tag("needs-verification")

# TODO:  This inadvertantly results in lspci being asked twice
#            if not has_file['lspci']:
#                reply.append_message(message_also_lspci_vvnn)
#                bug.append_tag("needs-lspci-vvnn")

# TODO:  Differentiate between crashes with and without full backtraces
#            if has_symptom['crash']:
#                reply.append_message(message_backtrace)

            # Try to identify packaging bugs
            # Subject =~ "depend on"|"wishlist"|"^Please"|"needs"

            if reply.get_message():
                count += 1
                total_count += 1

                if opt_dryrun:
                    continue

                text = ""
                try:
                    text = reply.make_reply()
                    bug.append_comment(text)
                    bugtask.transitionToStatus(status = new_status)
                    print bug.id, "  ", bugtask.status, "  ", bug.title
                    print " ---> " + new_status
                except:
                    print " ERROR:  Failed to update bug ", bug.id
                    print "Message was:  ", text
                    print "New status was:  ", new_status
                    raise

            elif has_file['backtrace'] and has_file['xorglog'] and has_file['lspci']:
                count += 1
                total_count += 1
                new_status = "Confirmed"
                bugtask.transitionToStatus(status = new_status)
                print bug.id, "  ", bugtask.status, "  ", bug.title
                print " --> Confirmed (has backtrace and log)"

            elif has_file['lspci'] and has_file['xorglog']:
                count += 1
                total_count += 1
                new_status = "Confirmed"
                bugtask.transitionToStatus(status = new_status)
                print bug.id, "  ", bugtask.status, "  ", bug.title
                print " --> Confirmed (has log and pci id info)"

            # TODO:  Need to test this in a dryrun first
            elif 1 == 2:
                # TODO:  Turn this into a subroutine
                for action in actions:
                    do_action = True
                    for f in actions['files']:
                        if not has_file[f]:
                            do_action = False
                            break
                    for nf in actions['no-files']:
                        if has_file[nf]:
                            do_action = False
                            break
                    for s in actions['symptoms']:
                        if not has_symptom[s]:
                            do_action = False
                            break
                    for ns in actions['non-symptoms']:
                        if has_symptom[ns]:
                            do_action = False
                            break

                    if do_action:
                        if action['status']:
                            print bug.id, "  ", bugtask.status, "  ", bug.title
                            print " --> %s (has %s)" % (action['status'], 'needed files') # TODO: list files
                        if action['reply']:
                            print bug.id, "  ", bugtask.status, "  ", bug.title
                            print "Would give reply here"  # TODO

        if count>0:
            print count," bugs updated for ",source_pkg

except:
    # TODO: lazr.restfulclient.errors.HTTPError: HTTP Error 500: Internal Server Error
    if count>0:
        print total_count," bugs updated"
    raise

if total_count>0:
    print total_count," bugs updated"
